// This file defines the global macros, enums and constants.

#ifndef MACROS_H
#define MACROS_H

#include "stl.h"

// Define if we want debug mode enabled. Warning: this might significantly slow
// down the assembler.
#define DEBUG 1

// Define whether to remove the ends of a contig. This is used in later stages
// of the assembler (e.g. geo-driver, merger).
#define CUT_TAILS 1

// Define if we want to run the assembler for single strand, in this case we
// will not consider the reverse compliment, if not defined, we will work with
// double strands. This mode of run is mainly defined for debugging purposes.
//#define SINGLE_STRAND 1

// This is for non-Harvard kind of datasets (e.g. Helicos) 
//#define NON_HARVARD 1

// Enable the use of quality score.
//#define USE_QUAL_SCORE 1

#define NO_UNKNOWN 1

// This is to enable compression for Helicos data in an attempt to get rid of
// homo-polymar problem.
//#define COMPRESS 1

// A big number
const int INF = 1000000000;

// # of mar types we use.
const int MAR_TYPES = 8;

// Define the mar types
enum mar_types                 {LMAR = 0, RMAR, LMARRC, RMARRC, LMARQ, RMARQ, LMARQRC, RMARQRC}; 

// Define the corresponding pair
const mar_types mar_type_pair [] = {RMAR, LMAR, RMARRC, LMARRC, RMARQ, LMARQ, RMARQRC, LMARQRC};

// Define bases.
enum base_types {BASEA = 0, BASET, BASEG, BASEC};

// How many bases should we use to hash a read/sequence
const int HASH_LEN = 6;

// Minimum length of a read.
const int MIN_READ_LEN = 10;

// Minimum # of bases to consider for overlap.
const int MIN_OVERLAP = 25;//15;//20;//15;//25;

// Minimum consequtive overlaps needed for an overlap. This is to prevent porous
// overlaps.
const int MIN_CONSEQ_OVERLAP = 7;

// Minimum score that will make an overlaxp eligible to merge.
const int MIN_SCORE = 75;//50;//75;

// This determines the amount of score that should come as a result of direct
// match. This restricts too many deletions/substitutions.
const int MIN_SCORE2OVERLAP = 70;//50;//75;

// This can be used to bypass the constraint set by 'MIN_OVERLAP'. This is
// useful for smaller reads, e.g. length < MIN_OVERLAP.
const int PERCENTAGE_LEN4OVERLAP = 80;

// Max lenght for which we can try overlap.
const int MAX_LEN4OVERLAP = 8000;

// Min length for which we allow deletion.
const int MIN_LEN4DELETION = 6;

// Percentage of deletion allowed.
const int PERCENTAGE_DEL_ALLOWED = 2;//5;//10;//2

// Co-efficient to be multiplied with 'match' scores.
const int COEFF_MATCH_SCORE = 2;

// Co-efficient to be multiplied with 'delete' scores.
const int COEFF_DEL_SCORE = 4;//5; //7

// Co-efficient to be multiplied with 'substitution' scores.
const int COEFF_SUBS_SCORE = 100;//100;//4;//100;//5; //8

// Specify the length which will qualify a contig to be used as a seed.
const int SEED_LEN = 30;//50; //100;  //30; 

const int KMAR_LEN = 8;

// # of paired reads. This is an upper bound.
const int NUM_PAIR = 500000;

// Average read length, an estimated value is ok.
const int AVG_READ_LEN = 25;

// # of deletions allowed to map a read in a seed. In future it can be
// percentage.
const int DEL_2MAP_SEED = 1;

// # of substitutions allowed to map a read in a seed. In future it can be
// percentage.
const int SUB_2MAP_SEED = 0;

// Average insert length of a paired read. Look below:
// LLLLLLL............RRRRRR
// <------INSERT_LEN------->
const int INSERT_LEN = 200; //1000-> for Harvard.

// Standard deviation in 'INSERT_LEN'.
const int INSERT_LEN_DEV = 100; //200;-> this was for Harvard

// This is a cut-off lenght for contigs. Anything below this will be ignored.
// This is sometimes used by the merger, vizualizer, etc.
const int CUT_OFF_LEN = 50;

// This is the percentage of how much we cut from the two ends of a contig. This
// is useful in the later stages when we want to merge long contigs but can't
// proceed due to tail errors.
const int PERCENTAGE_CUT_CONTIG_TAIL = 10;

// Prefix to add before all output files starting from root.
const string PREFIX_OUT = "out/";

#endif  //  MACROS_H
