// This will generate various statistics from a set of contigs and reference
// genome.

#include <fstream>
#include <iostream>
#include <function.h>  // functional didn't work :(

#include "header.h"
#include "kmar-table.h"
#include "utils-math.h"
#include "utils-string.h"

char tab = '\t';

int del_allowed;  // Cutoff for deletion error.
int ins_allowed;  // Cutoff for insertion error.

string org, orgrc;  // [Original, RC of original] sequence. 

typedef struct contig {
  bool rc;  // Did it map to the RC?
  int score, start;  // Score and the starting point.
  int homo_ins, homo_del;  // # of insertion/deletions that were in a homo-poly.
  double match, ins, del;  // # # of [match, ins, del] expressed as percentage.
  string str, aligned_str; // [Actual, aligned] sequence.
};

vector<contig> contigs;

IntVec bin;  // Define the bins for contig length distribution.
IntVec pgap; // Percentiles for gap length distribution.
IntVec plen; // Percentiles for contig length distribution.
IntVec acov; // Accuracy level of contigs for coverage.

// These are the tables used by kmar-tables. Declared here to make them global.
IntVec kmar_table[1 << (KMAR_LEN << 1)];
IntVec kmar_rc_table[1 << (KMAR_LEN << 1)];

void coverage_info(bool rc, int lower_len_limit, double upper_del_limit,
		   double upper_ins_limit);
void gap_info(bool rc, int lower_len_limit,double upper_del_limit,
	      double upper_ins_limit);
IntVec get_gaps(IntPairVec* x);
void length_info();
IntPairVec select_interval(bool rc, int lower_len_limit, double upper_del_limit,
			   double upper_ins_limit, double lower_acc_limit);

void init() {
  bin.push_back(0); bin.push_back(50); bin.push_back(100); bin.push_back(150);
  bin.push_back(200); bin.push_back(300); bin.push_back(500);
  bin.push_back(750); bin.push_back(1000); bin.push_back(1500);
  bin.push_back(2000); bin.push_back(2500); bin.push_back(3000);
  bin.push_back(5000);

  plen.push_back(10); plen.push_back(20); plen.push_back(30);
  plen.push_back(40); plen.push_back(50); plen.push_back(60);
  plen.push_back(70); plen.push_back(80); plen.push_back(90);
  plen.push_back(99);

  pgap.push_back(30); pgap.push_back(50); pgap.push_back(75);
  pgap.push_back(90); pgap.push_back(99);

  acov.push_back(98);
  acov.push_back(95);
  acov.push_back(90);
  acov.push_back(85);
}

bool cmp_len(const contig& a, const contig& b) {
  return a.str.size() < b.str.size();
}

bool cmp_interval(const IntPair& a, const IntPair& b) {
  return a.first < b.first;
}

void print_char(char c, int t, bool nl = false) {
  for (int i = 0; i < t; ++i) cout << c;
  if (nl) cout << endl;
}

int main(int argc, char* argv[]) {
  if (argc != 6) {
    cout << argc;
    cout << "usage: shorty-gen-stats org-file min-len ins del in-file" << endl;
    exit(1);
  }
  cout << "caution: did u give me the original org file?\n";
  init();
  int min_len = atoi(argv[2]);
  ins_allowed = atoi(argv[3]);
  del_allowed = atoi(argv[4]);
  fstream fin(argv[1], ios::in);
  fin >> org;
#ifdef COMPRESS
  org = UtilsString::CompressStr(org);
#endif  // COMPRESS
#ifdef COLOR_SPACE
  org = UtilsString::Fasta2Solid(org);
  orgrc = UtilsString::Reverse(org);
#endif  // COLOR_SPACE
#ifndef COLOR_SPACE
  orgrc = UtilsString::ReverseComplement(org);
#endif  // COLOR_SPACE
  int total_del = 0, total_ins = 0, total_homo_del = 0, total_homo_ins = 0;
  string str;
  fstream sin(argv[5], ios::in);
  KmarTable ktable(org, kmar_table), ktablerc(orgrc, kmar_rc_table);
  while (sin >> str) {
    if (str[0] == '>') continue;
    if (str.size() < min_len) continue; //temporary
    bool rc;
    int score, start, match, ins, homo_ins, del, homo_del;
    string aligned_str;
//     UtilsString::GetBestAlign(org, str, 0, org.length(), &rc, &score, &start,
// 			      &match, &ins, &homo_ins, &del, &homo_del,
// 			      &aligned_str);
    UtilsString::GetBestAlign(org, str, ktable, ktablerc, &rc, &score, &start,
			      &match, &ins, &homo_ins, &del, &homo_del,
			      &aligned_str);
    //    cout << "processed " << str << ' ' << match << endl;
    if (score <= 0) continue;
    contig t;
    t.str = str;
    t.rc = rc;
    t.score = score;
    t.start = start;
    t.match = 100.0 * match / str.size();
    t.ins = 100.0 * ins / str.size();

    // This is only for qualified contigs.
    if (t.ins <= ins_allowed) {
      total_ins += ins;
      total_homo_ins += homo_ins;
    }
    t.homo_ins = homo_ins;
    t.del = 100.0 * del / str.size();

    cout << "processed " << t.str << ' ' << t.match - t.ins - t.del << endl;

    // This is only for qualified contigs.
    if (t.del <= del_allowed) {
      total_del += del;
      total_homo_del += homo_del;
    }
    t.homo_del = homo_del;
    t.aligned_str = aligned_str;
    contigs.push_back(t);
  }
  cout << "INPUT FILE USED: " << argv[5] << endl;
  
  // Len info.
  // Print the header.
  print_char(' ', 50);
  cout << "CONTIG-LEN-INFO" << endl << endl;
  cout << "All contigs are considered" << endl;
  cout << "total" << "\taverage" << "\tsdev" << "\tmin" << "\tmax";
  for (int i = 0; i < plen.size(); ++i) {
    cout << tab << 'p' << plen[i];
    if (!i) cout << '*';
  }
  cout << endl;
  print_char('#', 120, true);
  length_info();
  print_char('#', 120, true);
  cout << "* 'p' stands for percentile" << endl;

  // Insertion/deletion info
  cout << endl;
  print_char(' ', 20);
  cout << "INDEL-INFO" << endl;
  print_char('#', 50, true);
  cout << "PERCENTAGE OF Insertion(s) ALLOWED: " << ins_allowed << endl;
  cout << "  " << 100.0 * total_homo_ins / total_ins
       << "% of insertions are in homo-polymer"<< endl;
  cout << "PERCENTAGE OF Deletion(s) ALLOWED: " << del_allowed << endl;
  cout << "  " << 100.0 * total_homo_del / total_del
       << "% of deletions are in homo-polymer"<< endl;
  print_char('#', 50, true);

  // Gap info.
  // Print the header.
  cout << endl;
  print_char(' ', 40);
  cout << "GAP-LEN-INFO" << endl << endl;
  cout << "Only contigs that survive the INDEL cutoff are considered" << endl;
  cout << "len*" << "\t#gaps" << "\taverage" << "\tsdev" << "\tmin" << "\tmax";
  for (int i = 0; i < pgap.size(); ++i) {
    cout << tab << 'p' << pgap[i];
    if (!i) cout << "**";
  }
  cout << endl;
  print_char('#', 85, true);
  char gap_sep[] = {'-', '-', '-', '-', '-', '-', '-', '#'};
  int gap_len[] = {6000, 3000, 2000, 1000, 300, 100, 50, 25};
  for (int i = 0; i < 8; ++i) {
    gap_info(false, gap_len[i], del_allowed, ins_allowed);
    gap_info(true, gap_len[i], del_allowed, ins_allowed);
    print_char(gap_sep[i], 85, true);
  }
  cout << "* this is the lower limit of length of contigs to be considered"
       << endl;
  cout << "** 'p' stands for percentile" << endl;

  // Coverage info.
  // Print the header.
  cout << endl;
  print_char(' ', 21);
  cout << "COVERAGE-INFO" << endl << endl;
  cout << "Only contigs that survive the INDEL cutoff are considered" << endl;
  cout << "len*" << "\t#contig";
  for (int i = 0; i < acov.size(); ++i) {
    cout << tab << 'a' << acov[i];
    if (!i) cout << "**";
  }
  cout << endl;
  print_char('#', 55, true);
  char cov_sep[] = {'-', '-', '-', '-', '-', '-', '-', '#'};
  int cov_len[] = {6000, 3000, 2000, 1000, 300, 100, 50, 25};
  for (int i = 0; i < 8; ++i) {
    coverage_info(false, cov_len[i], del_allowed, ins_allowed);
    coverage_info(true, cov_len[i], del_allowed, ins_allowed);
    print_char(cov_sep[i], 55, true);
  }
  cout << "* this is the lower limit of length of contigs to be considered"
       << endl;
  cout << "** 'aX' stands for the percentage of the original sequence that is "
    "covered by contigs with accuracy X%" << endl;

  return 0;
}

void coverage_info(bool rc, int lower_len_limit, double upper_del_limit,
		   double upper_ins_limit) {
  if (rc) cout << "  (RC)";
  else cout << lower_len_limit;
  cout.precision(5);
  cout << tab << select_interval(rc, lower_len_limit, upper_del_limit,
				 upper_ins_limit, 0).size();
  for (int i = 0; i < acov.size(); ++i) {
    IntPairVec x = select_interval(rc, lower_len_limit, upper_del_limit,
				   upper_ins_limit, acov[i]);
    x.push_back(IntPair(org.size(), org.size()));
    const IntVec& gaps = get_gaps(&x);
    double coverage = (org.size() - accumulate(gaps.begin(), gaps.end(), 0)) /
      (double)org.size();
    cout << tab << 100 * coverage;
  }
  cout << endl;
}

void gap_info(bool rc, int lower_len_limit, double upper_del_limit,
	      double upper_ins_limit) {
  if (rc) cout << "  (RC)";
  else cout << lower_len_limit;
  cout.precision(5);
  IntPairVec x;
  x = select_interval(rc, lower_len_limit, upper_del_limit, upper_ins_limit, 0);
  x.push_back(IntPair(org.size(), org.size()));
  IntVec gap = get_gaps(&x);
  cout << tab << gap.size();
  if (gap.size()) {
    cout << tab << UtilsMath::Average<IntVec>(gap) << tab
	 << UtilsMath::StandardDev<IntVec>(gap) << tab
	 << *(min_element(gap.begin(), gap.end())) << tab
	 << *(max_element(gap.begin(), gap.end()));
  } else {
    cout << tab << '-' << tab << '-' << tab << '-' << tab << '-';
  }
  for (int i = 0; i < pgap.size(); ++i) {
    if (gap.size()) {
      cout << tab << UtilsMath::IthPercentile<int, IntVec>(pgap[i], &gap);
    } else {
      cout << tab << '-';
    }
  }
  cout << endl;
}

IntVec get_gaps(IntPairVec* x) {
  IntVec gap;
  sort(x->begin(), x->end(), cmp_interval);
  int prev = -1;
  for (int i = 0; i < x->size(); ++i) {
    if ((*x)[i].first > prev) {  // This is a gap.
      if ((*x)[i].first - prev - 1) gap.push_back((*x)[i].first - prev - 1);
    }
    if ((*x)[i].second > prev) {
      prev = (*x)[i].second;
    }
  }
  return gap;
}

// There are many unnecessary use of stl functions here. I am not a professional
// developer, I am a student. So, I am trying to learn something.
void length_info() {
  IntVec len;
  for (int i = 0; i < contigs.size(); ++i) {
    len.push_back(contigs[i].str.size());
  }
  cout << len.size();
  if (len.size()) {
    cout << tab << UtilsMath::Average<IntVec>(len) << tab
	 << UtilsMath::StandardDev<IntVec>(len) << tab
	 << *(min_element(len.begin(), len.end())) << tab
	 << *(max_element(len.begin(), len.end()));
  } else {
    cout << tab << '-' << tab << '-' << tab << '-' << tab << '-';
  }
  for (int i = 0; i < plen.size(); ++i) {
    if (len.size()) {
      cout << tab << UtilsMath::IthPercentile<int, IntVec>(plen[i], &len);
    } else {
      cout << tab << '-';
    }
  }
  cout << endl << endl << "BINNING-INFO" << endl;
  sort(len.begin(), len.end());
  const IntVec& x = bin;
  for (int i = 1; i < x.size(); ++i) {
    cout << " [" << x[i - 1] << ", "<< x[i] << "):" << tab
	 << count_if(len.begin(), len.end(),
		     compose2(logical_and<bool>(), bind2nd(less<int>(), x[i]),
			      bind2nd(greater_equal<int>(), x[i - 1]))) << endl;
  }
  cout << " [" << x[x.size() - 1] << ", inf):" << tab
       << count_if(len.begin(), len.end(), 
		   bind2nd(greater_equal<int>(), x[x.size() - 1]))
       << endl;
}

IntPairVec select_interval(bool rc, int lower_len_limit, double upper_del_limit,
			   double upper_ins_limit, double lower_acc_limit) {
  IntPairVec x;
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].rc == rc && contigs[i].str.size() >= lower_len_limit &&
	contigs[i].del <= upper_del_limit &&
	contigs[i].ins <= upper_ins_limit &&
	contigs[i].match - contigs[i].ins - contigs[i].del >= lower_acc_limit) {
      x.push_back(IntPair(contigs[i].start, contigs[i].start
			  + contigs[i].aligned_str.size() - 1));
    }
  }
  return x;
}
