#ifndef OVERLAP_GRAPH_H
#define OVERLAP_GRAPH_H

#include "macros.h"
#include "shorty-string.h"
#include "header.h"

class Cluster;
class Contig;
class OverlapGraphNode;
class ReadProcessor;
class SeededRead;

// Builds a graph of contigs based on overlap information and provides
// methods to compress it to produce larger contigs.  
class OverlapGraph {

 public:
  // Create the graph from the cluster 'cluster'. 'log' will be used to log any
  // event that might occur.
  OverlapGraph(bool right, const Cluster& cluster, const SeededRead* seed,
	       const ReadProcessor* rp, fstream* log, fstream* contigs_out);
  
  // For the standalone version.
  OverlapGraph();

  ~OverlapGraph();

  // Needs to be called manually for the standalone version.
  void AllocateData();

  // Add a node to the graph with 'str', 'id', 'offset' & 'gen'. Assuming we are
  // not attempting to enter the same node twice, no error checking this time.
  // void AddNode(const ShortyString& str, int id, int offset, int so, int gen
  void AddNode(const ShortyString& mar, int id, int offset, int gen,
	       mar_types mt);

  // For the standalone version.
  void AddNode(const ShortyString& mar, int id, int gen);

  // Some of the reads were not used. Return them to read_processor_ for
  // future use.
  void ReUseReads();

  // Same as before but returns reads used by a contig with id 'x'.
  void ReUseReads(int x);

  // Print the contigs. If you call after merging, it will print the
  // merged contigs.
  //  void PrintContigs(vector<pair<SStrIntIntIntQuad, int> >* contigs);
  void PrintContigs(vector<Contig>* contigs);

  // This will construct an overlap graph and find long paths in order to
  // construct long contigs.
  void VisitGraph(vector<Contig>* contigs);

 private:

  // 'true' if it is right (in 5'-3' direction) to its seed, 'false' otherwise.
  bool right_;

  // 'true' if we are running a standalone version.
  bool standalone_;

  const SeededRead* seed_;

  class ContigInfo {
  public:
  ContigInfo(const ShortyString& str, int id, int offset, int gen,
	     mar_types mt) : str_(str), id_(id), offset_(offset), gen_(gen),
      mt_(mt) {}
    ShortyString str_;
    int id_, offset_, gen_;
    mar_types mt_;
    inline int size() const { return str_.size(); }
  };

  vector<ContigInfo> nodes_;

  // Adjacency list. Stores id, offset & weight of overlap.
  IntTripleVec* adj_;

  // Current path (id, offset) while traversing the graph.
  IntPair* curr_path_;

  // Mark whether a node is visited or not.
  bool* visited_;

  bool* path_valid_;

  // Longest string starting with the i'th read.
  ShortyString* lstring_;

  // Length of the longest contig starting with the i'th read.
  int* lclen_;

  // Store the path that forms the longest contig starting with the i'th read.
  IntPairVec* lpath_;

  // We store the read ids in buckets based on k-mer occurrences. This will make
  // the process of finding overlapping reads faster.
  set<int> buckets_[1 << (HASH_LEN << 1)];

  struct SortAdj {
    bool operator() (const IntTriple& a, const IntTriple& b) const {
      return a.third > b.third;
    }
  };

  struct IntPairEq {
    bool operator() (const IntPair& a, const IntPair& b) const {
      long long x = a.first * 1000000 + a.second;
      long long y = b.first * 1000000 + b.second;
      return x < y;
    }
  };
  set<IntPair, IntPairEq> unique_pairs_;

  // Helper method for VisitGraph.
  void BuildGraph();

  // Helper method for VisitGraph. elength-> expected length,
  // poffset-> previous offset.
  void Traverse(int root, int depth, int elength, int poffset);

  // Helper method for VisitGraph.
  void MergeReads(const IntPairVec& reads);

  // Pointer to a log file to record all events.
  fstream* log_;

  // Print the contig info.
  fstream* contigs_out_;

  const ReadProcessor* read_processor_;
};

#endif  // OVERLAP_GRAPH_H
