#include <iostream>

#include "shorty-string.h"
#include "utils-string.h"

using namespace std;

ShortyString::ShortyString() {
  str_ = "";
} 

ShortyString::ShortyString(const string& str, const VoteVec& votes)
  : str_(str), votes_(votes) {
#ifdef COMPRESS
  for (int i = 0; i < str.size(); ++i) {
    HomoPoly hp;
    hp[UtilsString::Char2Bt(str[i])] = 1;
    homo_poly_.push_back(hp);
  }
#endif  // COMPRESS
}

#ifdef COMPRESS
ShortyString::ShortyString(const string& str, const HomoPolyVec& hpv)
  : str_(str), homo_poly_(hpv) {
  for (int i = 0; i < str.size(); ++i) {
    Vote vote;
    vote[UtilsString::Char2Bt(str[i])] = 1;
    votes_.push_back(vote);
  }
}

ShortyString::ShortyString(const string& str, const HomoPolyVec& hpv,
			   const VoteVec& votes) 
  : str_(str), homo_poly_(hpv), votes_(votes) {
}
#endif  // COMPRESS

ShortyString::ShortyString(const string& str) {
  //  init();
  str_ = str;
  for (int i = 0; i < (int)str.size(); ++i) {
    Vote vote;
    vote.Inc(UtilsString::Char2Bt(str[i]), 1);
    votes_.push_back(vote);
#ifdef COMPRESS
    HomoPoly hp;
    hp[UtilsString::Char2Bt(str[i])] = 1;
    homo_poly_.push_back(hp);
#endif  // COMPRESS
  }
}

ShortyString::ShortyString(const string& str, int x) {
  str_ = str;
  for (int i = 0; i < (int)str.size(); ++i) {
    Vote vote;
    vote.Inc(UtilsString::Char2Bt(str[i]), x);
    votes_.push_back(vote);
#ifdef COMPRESS
    HomoPoly hp;
    hp[UtilsString::Char2Bt(str[i])] = 1;
    homo_poly_.push_back(hp);    
#endif  // COMPRESS
  }
}

ShortyString::~ShortyString() {
  VoteVec().swap(votes_);
}

void ShortyString::push_back(char c) {
  str_.push_back(c);
  Vote v;
  v.Inc(UtilsString::Char2Bt(c), 1);
  votes_.push_back(v);
#ifdef COMPRESS
  HomoPoly hp;
  hp[UtilsString::Char2Bt(c)] = 1;
  homo_poly_.push_back(hp);      
#endif  // COMPRESS
}

void ShortyString::push_back(char c, const Vote& v) {
  str_.push_back(c);
  votes_.push_back(v);
#ifdef COMPRESS
  HomoPoly hp;
  hp[UtilsString::Char2Bt(c)] = 1;
  homo_poly_.push_back(hp);      
#endif  // COMPRESS
}

#ifdef COMPRESS
void ShortyString::push_back(char c, const HomoPoly& hp, const Vote& v) {
  str_.push_back(c);
  votes_.push_back(v);
  homo_poly_.push_back(hp);
}
#endif  // COMPRESS

void ShortyString::reverse() { 
  std::reverse(str_.begin(), str_.end());
  std::reverse(votes_.begin(), votes_.end());
#ifdef COMPRESS
  std::reverse(homo_poly_.begin(), homo_poly_.end());
#endif  // COMPRESS
}

ShortyString ShortyString::substr(int pos, int len) const {
  assert(pos + len <= (int)str_.size());
  const string& str = str_.substr(pos, len);
  VoteVec votes;
  copy(votes_.begin() + pos, votes_.begin() + pos + len,
       back_insert_iterator<VoteVec>(votes));
#ifdef COMPRESS
  HomoPolyVec hpv;
  copy(homo_poly_.begin() + pos, homo_poly_.begin() + pos + len,
       back_insert_iterator<HomoPolyVec>(hpv));
  return ShortyString(str, hpv, votes);
#endif  // COMPRESS
  return ShortyString(str, votes);
}

int ShortyString::GetMaxVote() const {
  int m = 0;
  for (int i = 0; i < votes_.size(); ++i) {
    m = max(m, votes_[i].MaxVote());
  }
  return m;
}

ShortyString ShortyString::GetConfidentClone() const {
  ShortyString t;

  int left = 0;
  for (int i = 0; i < (int)str_.size(); ++i) {
    
    // Get the representative votes first. I think it is easy to avoid one
    // computataion by sliding the value from 'right' to 'my'. Will do it later.
    //    int left = i ? votes_[i - 1][UtilsString::Char2Bt(str_[i - 1])] : 0;
    int my = votes_[i][UtilsString::Char2Bt(str_[i])];
    int right =  (i < (int)str_.size() - 1) ?
      votes_[i + 1][UtilsString::Char2Bt(str_[i + 1])] : 0;
    if (100 * my >= BASE_CONFIDENCE * left && // || ??
	100 * my >= BASE_CONFIDENCE * right) {
#ifndef COMPRESS
      t.push_back(str_[i], votes_[i]);
#endif  // COMPRESS
#ifdef COMPRESS
      t.push_back(str_[i], homo_poly_[i], votes_[i]);
#endif  // COMPRESS
      left = my;
    }
  }

  t.reverse();
  ShortyString ret;
  left = 0;
  for (int i = 0; i < (int)t.Str().size(); ++i) {
    int my = t.Votes()[i][UtilsString::Char2Bt(t.Str()[i])];
    int right =  (i < (int)t.Str().size() - 1) ?
      t.Votes()[i + 1][UtilsString::Char2Bt(t.Str()[i + 1])] : 0;
    if (100 * my >= BASE_CONFIDENCE * left ||
	100 * my >= BASE_CONFIDENCE * right) {
#ifndef COMPRESS
     ret.push_back(t.Str()[i], t.Votes()[i]);
#endif  // COMPRESS
#ifdef COMPRESS
     ret.push_back(t.Str()[i], t.HomoPolies()[i], t.Votes()[i]);
#endif  // COMPRESS
      left = my;
    }
  }
  ret.reverse();
  return ret;
}

IntVec ShortyString::RepresentativeVotes() const {
  IntVec ret;
  for (int i = 0; i < (int)votes_.size(); ++i) {
    ret.push_back(votes_[i][UtilsString::Char2Bt(str_[i])]);
  }
  return ret;
}

void ShortyString::Print() const {
  for (int i = 0; i < (int)str_.size(); ++i) {
    cout << str_[i] << '[';
    for (int j = 0; j < BASE_TYPES; ++j) {
      if (j) cout << '.';
      cout << votes_[i][(base_types)j];
    }
    cout << ']';
  }
  cout << endl;
}

void ShortyString::PrintScores(ostream& fout) const {
  for (int i = 0; i < (int)str_.size(); ++i) {
    for (int j = 0; j < BASE_TYPES; ++j) {
      fout << votes_[i][(base_types)j] << ' ';
    }
  }
  fout << endl;
}

void ShortyString::ReadVotesFromFile(istream& fin) {
  for (int i = 0; i < (int)str_.size(); ++i) {
    fin >> votes_[i];
  }
}

ShortyString ShortyString::ReverseComplement() const {
  string str = str_;
  VoteVec votes = votes_;
#ifdef COMPRESS
  HomoPolyVec hpv = homo_poly_;
#endif  // COMPRESS
  for (int i = 0; i < (int)str.size(); ++i) {

    // Swap the votes and homo-polies first.
    swap(votes[i][UtilsString::Char2Bt(str[i])],
	 votes[i][UtilsString::Char2Bt(UtilsString::Complement(str[i]))]);
#ifdef COMPRESS
    swap(hpv[i][UtilsString::Char2Bt(str[i])],
	 hpv[i][UtilsString::Char2Bt(UtilsString::Complement(str[i]))]);    
#endif  // COMPRESS
    str[i] = UtilsString::Complement(str[i]);
  }
#ifndef COMPRESS
  ShortyString ret(str, votes);
#endif  // COMPRESS
#ifdef COMPRESS
  ShortyString ret(str, hpv, votes);
#endif 
 ret.reverse();
  return ret;
}

// Operator overloading.
ShortyString ShortyString::operator + (const ShortyString& ss) {
  ShortyString temp;
  temp.Str(Str() + ss.Str());

  // Combine the votes.
  VoteVec t;
  back_insert_iterator<VoteVec> i(t);
  copy(votes_.begin(), votes_.end(), i);
  copy(ss.Votes().begin(), ss.Votes().end(), i);
  temp.Votes(t);

  // Combine the homo-polies.
#ifdef COMPRESS
  HomoPolyVec hpv;
  back_insert_iterator<HomoPolyVec> j(hpv);
  copy(homo_poly_.begin(), homo_poly_.end(), j);
  copy(ss.HomoPolies().begin(), ss.HomoPolies().end(), j);
  temp.HomoPolies(hpv);
#endif  // COMPRESS

  return temp;
}

ShortyString ShortyString::operator + (const string& s) {
  ShortyString temp;
  const ShortyString& ss = ShortyString(s);
  temp.Str(Str() + ss.Str());

  // Combine the votes.
  VoteVec t;
  back_insert_iterator<VoteVec> i(t);
  copy(votes_.begin(), votes_.end(), i);
  copy(ss.Votes().begin(), ss.Votes().end(), i);
  temp.Votes(t);

  // Combine the homo-polies.
#ifdef COMPRESS
  HomoPolyVec hpv;
  back_insert_iterator<HomoPolyVec> j(hpv);
  copy(homo_poly_.begin(), homo_poly_.end(), j);
  copy(ss.HomoPolies().begin(), ss.HomoPolies().end(), j);
  temp.HomoPolies(hpv);
#endif  // COMPRESS

  return temp;
}

void ShortyString::operator += (const ShortyString& ss) {
  str_ += ss.Str();

  // Combine the votes.
  back_insert_iterator<VoteVec> i(votes_);
  copy(ss.Votes().begin(), ss.Votes().end(), i);

  // Combine the holo-polies.
#ifdef COMPRESS
  back_insert_iterator<HomoPolyVec> j(homo_poly_);
  copy(ss.HomoPolies().begin(), ss.HomoPolies().end(), j);
#endif  //COMPRESS
}

void ShortyString::operator += (const string& s) {
  const ShortyString& ss = ShortyString(s);
  str_ += ss.Str();

  // Combine the votes.
  back_insert_iterator<VoteVec> i(votes_);
  copy(ss.Votes().begin(), ss.Votes().end(), i);

  // Combine the homo-polies.
#ifdef COMPRESS
  back_insert_iterator<HomoPolyVec> j(homo_poly_);
  copy(ss.HomoPolies().begin(), ss.HomoPolies().end(), j);
#endif  // COMPRESS
}

void ShortyString::operator = (const string& str) {
  *this = ShortyString(str);
}

bool ShortyString::operator == (const ShortyString& ss) const {
  if (str_ != ss.Str()) return false;
  const VoteVec& v = ss.Votes();
  for (int i = 0; i < votes_.size(); ++i) {
    if (v[i] != votes_[i]) return false;
#ifdef COMPRESS
    // Doing nothing with homo-poly at this time.
#endif  // COMPRESS
  }
  return true;
}

bool ShortyString::operator == (const string& str) const {
  return str_ == str;
}

bool ShortyString::operator != (const string& str) const {
  return str_ != str;
}

char ShortyString::operator [] (int i) const {
  return str_[i];
}

char& ShortyString::operator [] (int i) {
  return str_[i];
}

istream& operator >> (istream& is, ShortyString& ss) {
  string str;
  is >> str;
  ss = ShortyString(str);
  return is;
}

ostream& operator << (ostream& os, const ShortyString& ss) { 
  os << ss.Str(); 
  return os;
}

// bool ShortyString::EquBaseCount(const BaseCount& a,const BaseCount& b) const {
//   return a[BASEA] == b[BASEA] && a[BASET] == b[BASET] && a[BASEG] == b[BASEG] &&
//     a[BASEC] == b[BASEC];
// }

// bool ShortyString::NEquBaseCount(const BaseCount& a, const BaseCount& b) const {
//   return a[BASEA] != b[BASEA] || a[BASET] != b[BASET] || a[BASEG] != b[BASEG] ||
//     a[BASEC] != b[BASEC];
// }
